#ifdef USE_AUDIO
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifdef USE_SDL1
#include <SDL/SDL.h>
#endif
#ifdef USE_SDL2
#include <SDL2/SDL.h>
#endif

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "config.h"
#include "const.h"
#include "nlllib.h"
#include "memory.h"
#include "audio.h"
#include "sound.h"
#include "sdl.h"

struct oper {
  struct oper *next;
  int index;
  int type;
  int freq;
  int phase;
  unsigned int flags;

  int outnum;
  struct {
    int amp;
    struct oper *op;
  } out[AUDIO_OPERATOR_OUTMAX];

  struct {
    int num;
    int amp;
  } in;

  struct {
    int num;
    struct envpoint {
      int step;
      int amp;
    } points[AUDIO_OPERATOR_ENVMAX];
  } envelope;
};

static struct opeset {
  int num;
  struct oper *opers;
} opesets[AUDIO_OPESET_MAXNUM];

struct operegs {
  struct operegs *next;

  struct oper *oper;

  struct {
    int amp;
  } out;

  struct {
    int index;
  } envelope;
};

struct samplestep {
  int sec;
  int count;
};

struct player {
  struct player *next;
  struct player *nextplayer;
  struct operegs *operegs;
  int amp;
  int scale;
  int octave;
  int freq;
  struct samplestep step;
  int steplen;
};

static struct player *playeres = NULL;

static struct oper *oper_pool = NULL;
static struct operegs *operegs_pool = NULL;
static struct player *player_pool = NULL;

static struct oper *_oper_alloc(void)
{
  struct oper *op;
  if (oper_pool) {
    op = oper_pool;
    oper_pool = oper_pool->next;
  } else {
    op = memory_alloc(sizeof(*op));
  }
  return op;
}

static void _oper_free(struct oper *op)
{
  op->next = oper_pool;
  oper_pool = op;
}

static struct operegs *_operegs_alloc(void)
{
  struct operegs *operegs;
  if (operegs_pool) {
    operegs = operegs_pool;
    operegs_pool = operegs_pool->next;
  } else {
    operegs = memory_alloc(sizeof(*operegs));
  }
  return operegs;
}

static void _operegs_free(struct operegs *operegs)
{
  operegs->next = operegs_pool;
  operegs_pool = operegs;
}

static struct player *_player_alloc(void)
{
  struct player *player;
  if (player_pool) {
    player = player_pool;
    player_pool = player_pool->next;
  } else {
    player = memory_alloc(sizeof(*player));
  }
  return player;
}

static void _player_free(struct player *player)
{
  player->next = player_pool;
  player_pool = player;
}

static int operegs_free(struct operegs *operegs)
{
  struct operegs *next;

  for (; operegs; operegs = next) {
    next = operegs->next;
    _operegs_free(operegs);
  }

  return 0;
}

static struct operegs *operegs_make(struct opeset *opeset)
{
  struct operegs *operegs = NULL, *regs;
  struct oper *op;

  for (op = opeset->opers; op; op = op->next) {
    if ((regs = _operegs_alloc()) == NULL)
      goto err;
    memset(regs, 0, sizeof(*regs));

    regs->oper = op;

    regs->next = operegs;
    operegs = regs;
  }

  return operegs;

err:
  operegs_free(operegs);
  return NULL;
}

static int step_clear(struct samplestep *step)
{
  step->sec   = 0;
  step->count = 0;
  return 0;
}

static int step_num(struct samplestep *step)
{
  return step->sec * AUDIO_SAMPLING_FREQ + step->count;
}

static int step_diff(struct samplestep *step0, struct samplestep *step1)
{
  return (step0->sec - step1->sec) * AUDIO_SAMPLING_FREQ + step0->count - step1->count;
}

static int step_copy(struct samplestep *dst, struct samplestep *src)
{
  dst->sec   = src->sec;
  dst->count = src->count;
  return 0;
}

static int step_add(struct samplestep *dst, struct samplestep *src, int count)
{
  struct samplestep step;

  if (!src)
    src = dst;

  step_copy(&step, src);

  step.count += count;
  while (step.count >= AUDIO_SAMPLING_FREQ) {
    step.sec++;
    step.count -= AUDIO_SAMPLING_FREQ;
  }

  step_copy(dst, &step);

  return 0;
}

#define SCALENUM 12

static int scale2freq[SCALENUM] = {
  16744036, /* C9  */
  17739688, /* C#9 */
  18794545, /* D9  */
  19912127, /* D#9 */
  21096164, /* E9  */
  22350607, /* F9  */
  23679643, /* F#9 */
  25087708, /* G9  */
  26579501, /* G#9 */
  28160000, /* A9  */
  29834481, /* A#9 */
  31608531, /* B9  */
};

static struct player *player_alloc(struct opeset *opeset,
				   int amp, int scale, int octave,
				   struct samplestep *step, int steplen)
{
  struct player *player;
  int freq = 0, i;

  if ((player = _player_alloc()) == NULL)
    return NULL;
  memset(player, 0, sizeof(*player));

  if (scale >= 0) {
    freq = scale2freq[scale];
    for (i = octave; i < 9; i++)
      freq >>= 1;
    freq /= 1000;
  }

  player->amp      = amp;
  player->scale    = scale;
  player->octave   = octave;
  player->freq     = freq;
  step_copy(&player->step, step);
  player->steplen  = steplen;

  player->operegs = operegs_make(opeset);

  return player;
}

static int player_free(struct player *player)
{
  struct player *next;

  for (; player; player = next) {
    if (player->nextplayer) {
      player_free(player->nextplayer);
      player->nextplayer = NULL;
    }
    next = player->next;
    operegs_free(player->operegs);
    _player_free(player);
  }

  return 0;
}

static int player_add(struct player **pp, struct opeset *opeset,
		      int amp, int scale, int octave,
		      struct samplestep *step, int steplen)
{
  struct player *player;

  player = player_alloc(opeset, amp, scale, octave, step, steplen);
  if (!player)
    return -1;

  player->next = *pp;
  *pp = player;

  return 0;
}

static int init(void)
{
  static int initialized = 0;

  if (!initialized) {
    memset(opesets, 0, sizeof(opesets));
#ifdef USE_SDL
    sdl_start();
#endif
    initialized = 1;
  }

  return 0;
}

int audio_check(void)
{
  int i;

  if (playeres)
    return NLL_ERRCODE_MEMORY_NOT_EMPTY;

  for (i = 0; i < AUDIO_OPESET_MAXNUM; i++) {
    if (opesets[i].opers || opesets[i].num)
      return NLL_ERRCODE_MEMORY_NOT_EMPTY;
  }

  return 0;
}

int audio_init(void)
{
  return 0;
}

int audio_done(void)
{
  int r;
  struct oper *oper;
  struct operegs *operegs;
  struct player *player;

  if ((r = player_free(playeres)) < 0)
    return r;
  playeres = NULL;

  if ((r = audio_clear_all()) < 0)
    return r;

  while (oper_pool) {
    oper = oper_pool;
    oper_pool = oper_pool->next;
    memory_free(oper);
  }

  while (operegs_pool) {
    operegs = operegs_pool;
    operegs_pool = operegs_pool->next;
    memory_free(operegs);
  }

  while (player_pool) {
    player = player_pool;
    player_pool = player_pool->next;
    memory_free(player);
  }

  return 0;
}

#define SINNUM 64

static int sin_val[SINNUM + 1] = {
  0,     245,  491,  736,  980, 1224, 1467, 1710,
  1951, 2191, 2430, 2667, 2903, 3137, 3369, 3599,
  3827, 4052, 4276, 4496, 4714, 4929, 5141, 5350,
  5556, 5758, 5957, 6152, 6344, 6532, 6716, 6895,
  7071, 7242, 7410, 7572, 7730, 7883, 8032, 8176,
  8315, 8449, 8577, 8701, 8819, 8932, 9040, 9142,
  9239, 9330, 9415, 9495, 9569, 9638, 9700, 9757,
  9808, 9853, 9892, 9925, 9952, 9973, 9988, 9997, 10000
};

static int degree_inrange(int deg)
{
  while (deg < 0)
    deg += (SINNUM*4);
  deg %= (SINNUM*4);
  return deg;
}

static int int_sin(int deg)
{
  deg = degree_inrange(deg);
  if (deg > (SINNUM*2))
    return -int_sin(deg - (SINNUM*2));
  if (deg > SINNUM)
    return int_sin((SINNUM*2) - deg);
  return sin_val[deg];
}

static int int_triangle(int deg)
{
  deg = degree_inrange(deg);
  if (deg > (SINNUM*2))
    return -int_triangle(deg - (SINNUM*2));
  if (deg > SINNUM)
    return int_triangle((SINNUM*2) - deg);
  return deg * 10000 / SINNUM;
}

static int int_rectangle(int deg)
{
  deg = degree_inrange(deg);
  return (deg < (SINNUM*2)) ? 10000 : -10000;
}

static int int_sawup(int deg)
{
  deg = degree_inrange(deg);
  return (deg - (SINNUM*2)) * 10000 / (SINNUM*2);
}

static int int_sawdown(int deg)
{
  deg = degree_inrange(deg);
  return ((SINNUM*2) - deg) * 10000 / (SINNUM*2);
}

static unsigned int seed = 1;

static unsigned int _rand(void)
{
  seed ^= seed << 13;
  seed ^= seed >> 17;
  seed ^= seed <<  5;
  return seed;
}

static int int_noise(int deg)
{
  return (_rand() % 20001) - 10000;
}

struct audio_data {
  volatile int finished;
  int num;
  int samplefreq;
  int channels;
  struct samplestep step;
  struct player *ready;
  struct player *busy;
  struct player *complete;
};

static int int_amp(int type, int deg)
{
  int a;
  switch (type) {
  case NLL_A_SIN:       a = int_sin(deg); break;
  case NLL_A_TRIANGLE:  a = int_triangle(deg); break;
  case NLL_A_RECTANGLE: a = int_rectangle(deg); break;
  case NLL_A_SAWUP:     a = int_sawup(deg); break;
  case NLL_A_SAWDOWN:   a = int_sawdown(deg); break;
  case NLL_A_NOISE:     a = int_noise(deg); break;
  default:              a = 0; break;
  }
  return a;
}

static struct oper *opeset_search(struct opeset *opeset, int index)
{
  struct oper *op;

  for (op = opeset->opers; op; op = op->next) {
    if (op->index == index)
      break;
  }

  return op;
}

static int opereg_calc_envelope(struct operegs *regs, int step)
{
  struct oper *op;
  struct envpoint *p0;
  struct envpoint *p1;

  op = regs->oper;

  if (!op->envelope.num)
    return 100;

  p0 = &op->envelope.points[regs->envelope.index];

  if (regs->envelope.index >= op->envelope.num - 1)
    return p0->amp;

  p1 = p0 + 1;

  if (p0->step >= p1->step)
    return p0->amp;

  return p0->amp + (p1->amp - p0->amp) * (step - p0->step) / (p1->step - p0->step);
}

static int calc_deg(int step, int freq, int samplefreq)
{
  int deg, d, m;

#if 0
  deg = step * freq * (SINNUM*4) / samplefreq;
#else /* To avoid overflow, calculate division and modulo separately */
  d = (step * freq) / samplefreq;
  m = (step * freq) % samplefreq;
  deg  = d * (SINNUM*4);
  deg += m * (SINNUM*4) / samplefreq;
#endif

  return deg;
}

static int opereg_calc(struct operegs *operegs, int amp, int freq,
		       int samplefreq, int step, int steplen)
{
  struct operegs *regs;
  struct oper *op;
  int i, deg, a, count = 0, outamp = 0, envstep;

  if (!operegs) {
    deg = calc_deg(step, freq, samplefreq);
    count = 1;
    outamp = int_amp(NLL_A_SIN, deg);
  }

  for (regs = operegs; regs; regs = regs->next) {
    op = regs->oper;
    op->in.num = 0;
    op->in.amp = 0;
  }

  for (regs = operegs; regs; regs = regs->next) {
    op = regs->oper;
    for (i = 0; i < op->outnum; i++) {
      if (op->out[i].op) {
	op->out[i].op->in.num++;
	op->out[i].op->in.amp += regs->out.amp * op->out[i].amp / 100;
      } else {
	count++;
	outamp += regs->out.amp * op->out[i].amp / 100;
      }
    }
  }

#if 1
  envstep = (steplen > 0) ? (step * 100 / steplen) : 0;
#else
  envstep = step * 100 / (samplefreq / 4);
#endif

  for (regs = operegs; regs; regs = regs->next) {
    op = regs->oper;
    deg = calc_deg(step, freq, samplefreq);
    deg = deg * op->freq / 100;
    deg += op->phase * (SINNUM*4) / 100;

    if (op->in.num)
      op->in.amp /= op->in.num;

    switch (op->flags & NLL_A_MOD_MASK) {
    case NLL_A_PHASE:
      deg = deg + op->in.amp * (SINNUM*2) / 10000;
      break;
    default:
      break;
    }

    a = int_amp(op->type, deg);

    switch (op->flags & NLL_A_MOD_MASK) {
    case NLL_A_NULL: a  =  op->in.amp; break;
    case NLL_A_REV:  a  = -op->in.amp; break;
    case NLL_A_SIGN: a  = (a < op->in.amp) ? -10000 : ((a > op->in.amp) ? 10000 : 0); break;
    case NLL_A_ADD:  a += op->in.amp; break;
    case NLL_A_SUB:  a -= op->in.amp; break;
    case NLL_A_MUL:  a *= op->in.amp; break;
    case NLL_A_AND:  a &= op->in.amp; break;
    case NLL_A_OR:   a |= op->in.amp; break;
    case NLL_A_XOR:  a ^= op->in.amp; break;
    case NLL_A_AVE:  a  = (a + op->in.amp) / 2; break;
    case NLL_A_PER:  a  = ((a / 10) * (op->in.amp / 10)) / 100; break;
    case NLL_A_MIN:  a  = (a < op->in.amp) ? a : op->in.amp; break;
    case NLL_A_MAX:  a  = (a > op->in.amp) ? a : op->in.amp; break;
    case NLL_A_LT:   a  = (a < op->in.amp) ? 10000 : -10000; break;
    case NLL_A_GT:   a  = (a > op->in.amp) ? 10000 : -10000; break;
    case NLL_A_PHASE:
    default:
      break;
    }

    regs->out.amp = a;

    while ((regs->envelope.index + 1 < op->envelope.num) &&
	   (op->envelope.points[regs->envelope.index + 1].step <= envstep)) {
      regs->envelope.index++;
    }
    regs->out.amp = regs->out.amp * opereg_calc_envelope(regs, envstep) / 100;
  }

  if (count)
    outamp /= count;

  return outamp * amp / 10; /* in range of S16 (signed short)*/
}

static int calc_frames(struct audio_data *audiodata, short *frames, int n)
{
  int i, amp;
  struct player *player, **pp;
  struct samplestep stepend;

  if (audiodata->finished) {
    for (i = 0; i < n; i++) {
      frames[i] = 0;
      step_add(&audiodata->step, NULL, 1);
    }
    return n;
  }

  step_add(&stepend, &audiodata->step, n);

  for (pp = &audiodata->ready; *pp;) {
    player = *pp;

    if (step_diff(&stepend, &player->step) < 0) {
      pp = &player->nextplayer;
      continue;
    }

    if (player->next) {
      *pp = player->next;
      (*pp)->nextplayer = player->nextplayer;
    } else {
      *pp = player->nextplayer;
    }

    player->nextplayer = NULL;

    player->next = audiodata->busy;
    audiodata->busy = player;
  }

  for (i = 0; i < n; i++) {
    amp = 0;
    for (player = audiodata->busy; player; player = player->next) {
      step_add(&stepend, &player->step, player->steplen);
      if ((step_diff(&audiodata->step, &player->step) < 0) ||
	  !(step_diff(&audiodata->step, &stepend) < 0))
	continue;
      amp += opereg_calc(player->operegs,
			 player->amp * 32 / 100, /* in range of S16 (signed short)*/
			 player->freq, audiodata->samplefreq,
			 step_diff(&audiodata->step, &player->step),
			 player->steplen);
    }
    amp = audiodata->num ? (amp / audiodata->num) : 0;
    if (amp >  32767) amp =  32767;
    if (amp < -32767) amp = -32767;
    frames[i] = amp;
    step_add(&audiodata->step, NULL, 1);
  }

  n = i;

  for (pp = &audiodata->busy; *pp;) {
    player = *pp;

    step_add(&stepend, &player->step, player->steplen);
    if (step_diff(&audiodata->step, &stepend) < 0) {
      pp = &player->next;
      continue;
    }

    *pp = player->next;

    player->next = audiodata->complete;
    audiodata->complete = player;
  }

  if (!audiodata->ready && !audiodata->busy)
    audiodata->finished = 1;

  return n;
}

#ifdef USE_SDL
static void sdl_callback(void *userdata, unsigned char *stream, int len)
{
  struct audio_data *audiodata;
  short *frames;
  int n;

  audiodata = (struct audio_data *)userdata;

  frames = (short *)stream;
  n = len / 2;

  calc_frames(audiodata, frames, n);
}
#endif

static int _run(struct audio_data *audiodata, struct samplestep *step)
{
#ifdef USE_SDL
  SDL_AudioSpec spec;
#endif

#ifdef USE_SDL
  memset(&spec, 0, sizeof(spec));

  spec.freq = audiodata->samplefreq;
  spec.format = AUDIO_S16;
  spec.channels = audiodata->channels;
  spec.samples = 256;
  spec.callback = sdl_callback;
  spec.userdata = audiodata;

  SDL_OpenAudio(&spec, NULL);
  SDL_PauseAudio(0);
#if 0
  SDL_Delay((step->sec * 1000) + (step->count * 1000 / audiodata->samplefreq));
#else
  while (!audiodata->finished) {
    SDL_Delay(10);
  }
#endif
  SDL_PauseAudio(1);
  SDL_CloseAudio();
#endif

  return 0;
}

static int _run_file(struct audio_data *audiodata, struct samplestep *step, const char *filename)
{
  FILE *fp;
  int r, bits = 16, offset = 0, n, samplenum;
#define FRAMENUM 256
  short frames[FRAMENUM];

  samplenum = step_num(step);

  fp = sound_open(filename, audiodata->samplefreq, audiodata->channels, bits, samplenum);
  if (!fp)
    return -1;

  for (offset = 0; offset < samplenum;) {
    n = FRAMENUM;
    if (offset + n >= samplenum)
      n = samplenum - offset;
    calc_frames(audiodata, frames, n);
    sound_write(fp, audiodata->channels, bits, samplenum, &offset, frames, n);
    if (audiodata->finished)
      break;
  }

  if ((r = sound_close(fp, audiodata->channels, bits, samplenum, &offset)) < 0)
    return -1;

  return 0;
}

static struct player *player_run(struct player *player, struct samplestep *step, const char *filename)
{
  static struct audio_data audiodata;
  struct player *p;
  int num = 0;

  init();

  for (p = player; p; p = p->nextplayer)
    num++;

  audiodata.finished = 0;
  audiodata.num = num;
  audiodata.samplefreq = AUDIO_SAMPLING_FREQ;
  audiodata.channels = 1;
  step_clear(&audiodata.step);

  audiodata.ready    = player;
  audiodata.busy     = NULL;
  audiodata.complete = NULL;

  if (filename)
    _run_file(&audiodata, step, filename);
  else
    _run(&audiodata, step);

  audiodata.finished = 1;

  while (audiodata.ready) {
    player = audiodata.ready;
    audiodata.ready = audiodata.ready->nextplayer;

    player->nextplayer = audiodata.complete;
    audiodata.complete = player;
  }

  while (audiodata.busy) {
    player = audiodata.busy;
    audiodata.busy = audiodata.busy->nextplayer;

    player->nextplayer = audiodata.complete;
    audiodata.complete = player;
  }

  return audiodata.complete;
}

int audio_clear(int opeset_index)
{
  struct opeset *opeset;
  struct oper *next;

  init();

  if (opeset_index >= AUDIO_OPESET_MAXNUM)
    return -1;

  if (opeset_index < 0)
    return audio_clear_all();

  opeset = &opesets[opeset_index];

  player_free(playeres);
  playeres = NULL;

  for (; opeset->opers; opeset->opers = next) {
    next = opeset->opers->next;
    _oper_free(opeset->opers);
  }

  opeset->num = 0;

  return 0;
}

int audio_clear_all(void)
{
  int r, i;

  init();

  for (i = 0; i < AUDIO_OPESET_MAXNUM; i++) {
    if ((r = audio_clear(i)) < 0)
      return r;
  }

  return 0;
}

int audio_operator(int output, int amp, int freq, int phase, int type, int flags, int opeset_index)
{
  struct opeset *opeset;
  struct oper *op;
  int r;

  init();

  if ((opeset_index < 0) || (opeset_index >= AUDIO_OPESET_MAXNUM))
    return -1;

  opeset = &opesets[opeset_index];

  if ((op = _oper_alloc()) == NULL)
    return -1;
  memset(op, 0, sizeof(*op));

  op->index = ++opeset->num;
  op->type = type;
  op->freq = freq;
  op->phase = phase;
  op->flags = flags;

  op->next = opeset->opers;
  opeset->opers = op;

  if (op->flags & NLL_A_CLEARENV) {
    op->envelope.num = 0;
  } else {
    op->envelope.num = 2;
    op->envelope.points[0].step = 0;
    op->envelope.points[0].amp  = 100;
    op->envelope.points[1].step = 100;
    op->envelope.points[1].amp  = 0;
  }

  if ((r = audio_addout(op->index, output, amp, opeset_index)) < 0) {
    opeset->opers = op->next;
    _oper_free(op);
    return r;
  }

  return op->index;
}

int audio_addout(int index, int output, int amp, int opeset_index)
{
  struct opeset *opeset;
  struct oper *op, *out;

  init();

  if ((opeset_index < 0) || (opeset_index >= AUDIO_OPESET_MAXNUM))
    return -1;

  opeset = &opesets[opeset_index];

  op = opeset_search(opeset, index);
  if (!op)
    return -1;

  if (op->outnum == AUDIO_OPERATOR_OUTMAX)
    return -1;

  if (output == NLL_A_OUTPUT) {
    out = NULL;
  } else {
    out = opeset_search(opeset, output);
    if (!out)
      return -1;
  }

  op->out[op->outnum  ].amp = amp;
  op->out[op->outnum++].op  = out;

  return 0;
}

int audio_envpoint(int index, int step, int amp, int opeset_index)
{
  struct opeset *opeset;
  struct oper *op;

  init();

  if ((opeset_index < 0) || (opeset_index >= AUDIO_OPESET_MAXNUM))
    return -1;

  opeset = &opesets[opeset_index];

  op = opeset_search(opeset, index);
  if (!op)
    return -1;

  if (op->envelope.num == AUDIO_OPERATOR_ENVMAX)
    return -1;

  op->envelope.points[op->envelope.num  ].step = step;
  op->envelope.points[op->envelope.num++].amp  = amp;

  return 0;
}

static int code2scale[] = { 0, 2, 4, 5, 7, 9, 11 };

int audio_play(const char *string, int flags, int opeset_index, const char *filename)
{
  struct opeset *opeset;
  static struct player **pp;
  int scale, octave = 3, oct, len;
  int interval, steplen;
  static int length = 4, amp = 70, tempo = 120, stepmod = 0;
  static struct samplestep step, max;
  int bracket = 0, bracketlen = 0;
  const char *s, *upper, *lower;
  char *p;

  init();

  if ((opeset_index < 0) || (opeset_index >= AUDIO_OPESET_MAXNUM))
    return -1;

  opeset = &opesets[opeset_index];

  upper = "CDEFGAB";
  lower = "cdefgab";

  if (!playeres) {
    step_clear(&max);
  }

  if (!playeres || (flags & NLL_A_NEW)) {
    pp = &playeres;
    step_clear(&step);
    stepmod = 0;
  }

  for (s = string; *s;) {
    switch (*s) {
    case ' ':
    case '\t':
    case '\r':
    case '\n':
      s++;
      break;

    case 'O':
      s++;
      if (*s && isdigit(*s)) {
	octave = *s - '0';
	s++;
      }
      break;

    case '<':
      s++;
      octave--;
      break;

    case '>':
      s++;
      octave++;
      break;

    case 'L':
      s++;
      if (*s && isdigit(*s)) {
	length = strtol(s, &p, 10);
	if (s == p)
	  goto err;
	s = p;
	length = length ? length : 4;
      }
      break;

    case 'V':
      s++;
      if (*s && isdigit(*s)) {
	amp = strtol(s, &p, 10);
	if (s == p)
	  goto err;
	s = p;
	amp = amp ? amp : 70;
      }
      break;

    case 'T':
      s++;
      if (*s && isdigit(*s)) {
	tempo = strtol(s, &p, 10);
	if (s == p)
	  goto err;
	s = p;
	tempo = tempo ? tempo : 120;
      }
      break;

    case '(':
      s++;
      if (bracket)
	goto err;
      bracketlen = 0;
      bracket = 1;
      break;

    case ')':
      s++;
      if (!bracket)
	goto err;
      step_add(&step, NULL, bracketlen);
      bracket = 0;
      break;

    case 'R':
    case 'r':
    default:
      if ((*s == 'R') || (*s == 'r')) {
	scale = -1;
      } else if ((p = strchr(upper, *s)) != NULL) {
	scale = code2scale[p - upper];
      } else if ((p = strchr(lower, *s)) != NULL) {
	scale = code2scale[p - lower];
      } else {
	goto err;
      }
      s++;

      if (scale < 0) {
	oct = -1;
      } else {
	oct = octave;

	if (*s && ((*s == '#') || (*s == '+'))) {
	  scale++;
	  s++;
	}

	if (*s && (*s == '-')) {
	  scale--;
	  s++;
	}

	while (scale < 0) {
	  scale += SCALENUM;
	  oct--;
	}

	while (scale >= SCALENUM) {
	  scale -= SCALENUM;
	  oct++;
	}
      }

      len = length;
      if (*s && isdigit(*s)) {
	len = strtol(s, &p, 10);
	if (s == p)
	  goto err;
	s = p;
      }

      if (len <= 0)
	goto err;
      interval = 64 / len;

      if (*s && (*s == '.')) {
	interval = interval * 3 / 2;
	s++;
      }

#if 0
      msec = (1000 * 60 / 16) * interval / tempo;
      steplen = msec * AUDIO_SAMPLING_FREQ / 1000;
#else /* To avoid the accumulation of the error, save the remainder */
      steplen = 60 * interval * AUDIO_SAMPLING_FREQ / 16;
      steplen += stepmod; /* add the old remainder */
      stepmod = steplen % tempo; /* save the remainder */
      steplen = steplen / tempo;
#endif

      if (player_add(pp, opeset, amp, scale, oct, &step, steplen) < 0)
	goto err;
      if (pp == &playeres) {
	(*pp)->nextplayer = (*pp)->next;
	(*pp)->next = NULL;
      }
      pp = &(*pp)->next;

      if (!bracket) {
	step_add(&step, NULL, steplen);
      } else {
	if (bracketlen < steplen)
	  bracketlen = steplen;
      }

      break;
    }
  }

  if (step_diff(&max, &step) < 0)
    step_copy(&max, &step);

  if (!(flags & NLL_A_SAVE)) {
    if (playeres) {
      playeres = player_run(playeres, &max, filename);
      if (!playeres)
	return -1;
      player_free(playeres);
      playeres = NULL;
    }
  }

  return 0;

err:
  return -1;
}

#endif
