#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/types.h>

#include "config.h"
#include "sound.h"

struct chunk_header {
#define CHUNK_HEADER_ID_RIFF 0x46464952 /* RIFF */
#define CHUNK_HEADER_ID_FMT  0x20746d66 /* fmt */
#define CHUNK_HEADER_ID_DATA 0x61746164 /* data */
  uint32_t id;
  uint32_t size;
};
#define CHUNK_HEADER_SIZE sizeof(struct chunk_header)

struct riff_chunk_header {
  struct chunk_header header;
#define RIFF_CHUNK_HEADER_FORMAT_WAVE 0x45564157 /* WAVE */
  uint32_t format;
};
#define RIFF_CHUNK_HEADER_SIZE sizeof(struct riff_chunk_header)

struct fmt_chunk_header {
  struct chunk_header header;
#define FMT_CHUNK_HEADER_FORMAT_TAG_PCM 0x0001
  uint16_t format_tag;
  uint16_t channels;
  uint32_t sample_per_sec;
  uint32_t avg_bytes_per_sec;
  uint16_t block_align;
  uint16_t bits_per_sample;
};
#define FMT_CHUNK_HEADER_SIZE sizeof(struct fmt_chunk_header)

struct data_chunk_header {
  struct chunk_header header;
};
#define DATA_CHUNK_HEADER_SIZE sizeof(struct data_chunk_header)

union evalue {
  uint8_t  b[4];
  uint16_t h[2];
  uint32_t w[1];
};

#if 0
static uint16_t read_le16(uint16_t val)
{
  union evalue e;
  e.h[0] = val;
  return (e.b[1] << 8) | e.b[0];
}

static uint32_t read_le32(uint32_t val)
{
  union evalue e;
  e.w[0] = val;
  return (e.b[3] << 24) | (e.b[2] << 16) | (e.b[1] << 8) | e.b[0];
}
#endif

static uint16_t write_le16(uint16_t val)
{
  union evalue e;
  e.b[0] =  val        & 0xff;
  e.b[1] = (val >>  8) & 0xff;
  return e.h[0];
}

static uint32_t write_le32(uint32_t val)
{
  union evalue e;
  e.b[0] =  val        & 0xff;
  e.b[1] = (val >>  8) & 0xff;
  e.b[2] = (val >> 16) & 0xff;
  e.b[3] = (val >> 24) & 0xff;
  return e.w[0];
}

FILE *sound_wave_open(const char *filename, int samplefreq, int channels, int bits,
		      int samplenum)
{
  FILE *fp = NULL;
  struct riff_chunk_header rhdr;
  struct fmt_chunk_header fhdr;
  struct data_chunk_header dhdr;
  int riffsize, fmtsize, datasize;

  if ((samplenum < 0) || (samplefreq <= 0))
    goto err;

  if ((channels != 1) && (channels != 2))
    goto err;

  if ((bits != 8) && (bits != 16) && (bits != 32))
    goto err;

  fp = fopen(filename, "w");
  if (!fp)
    goto err;

  datasize = DATA_CHUNK_HEADER_SIZE + (channels * (bits / 8) * samplenum);
  fmtsize  = FMT_CHUNK_HEADER_SIZE;
  riffsize = RIFF_CHUNK_HEADER_SIZE + fmtsize + datasize;

  memset(&rhdr, 0, sizeof(rhdr));
  memset(&fhdr, 0, sizeof(fhdr));
  memset(&dhdr, 0, sizeof(dhdr));

  rhdr.header.id         = write_le32(CHUNK_HEADER_ID_RIFF);
  rhdr.header.size       = write_le32(riffsize - CHUNK_HEADER_SIZE);
  rhdr.format            = write_le32(RIFF_CHUNK_HEADER_FORMAT_WAVE);

  fhdr.header.id         = write_le32(CHUNK_HEADER_ID_FMT);
  fhdr.header.size       = write_le32(fmtsize - CHUNK_HEADER_SIZE);
  fhdr.format_tag        = write_le16(FMT_CHUNK_HEADER_FORMAT_TAG_PCM);
  fhdr.channels          = write_le16(channels);
  fhdr.sample_per_sec    = write_le32(samplefreq);
  fhdr.avg_bytes_per_sec = write_le32((bits / 8) * channels * samplefreq);
  fhdr.block_align       = write_le16((bits / 8) * channels);
  fhdr.bits_per_sample   = write_le16(bits);

  dhdr.header.id         = write_le32(CHUNK_HEADER_ID_DATA);
  dhdr.header.size       = write_le32(datasize - CHUNK_HEADER_SIZE);

  if (fwrite(&rhdr, 1, RIFF_CHUNK_HEADER_SIZE, fp) != RIFF_CHUNK_HEADER_SIZE)
    goto err;

  if (fwrite(&fhdr, 1, FMT_CHUNK_HEADER_SIZE, fp) != FMT_CHUNK_HEADER_SIZE)
    goto err;

  if (fwrite(&dhdr, 1, DATA_CHUNK_HEADER_SIZE, fp) != DATA_CHUNK_HEADER_SIZE)
    goto err;

  return fp;

err:
  if (fp)
    fclose(fp);
  return NULL;
}

int sound_wave_write(FILE *fp, int channels, int bits, int samplenum,
		     int *offset, void *frames, int num)
{
  int i, ch, n = 0;
  union evalue e;

  if (!fp)
    return -1;

  for (i = 0; i < num; i++) {
    for (ch = 0; ch < channels; ch++) {
      memset(&e, 0, sizeof(e));
      if (*offset < samplenum) {
	switch (bits) {
	case  8: e.b[0] =            ((int8_t  *)frames)[n] ; break;
	case 16: e.h[0] = write_le16(((int16_t *)frames)[n]); break;
	case 32: e.w[0] = write_le32(((int32_t *)frames)[n]); break;
	default: break;
	}
      }
      fwrite(&e, 1, bits / 8, fp);
      n++;
    }
    (*offset)++;
  }

  return 0;
}

int sound_wave_close(FILE *fp, int channels, int bits, int samplenum,
		     int *offset)
{
  int align_num, ch;
  union evalue e;

  if (!fp)
    return -1;

  align_num = samplenum;
  if ((channels == 1) && (bits == 8) && ((samplenum % 2) != 0))
    align_num++;

  memset(&e, 0, sizeof(e));

  while (*offset < align_num) {
    for (ch = 0; ch < channels; ch++) {
      fwrite(&e, 1, bits / 8, fp);
    }
    (*offset)++;
  }

  fclose(fp);

  return 0;
}

FILE *sound_open(const char *filename, int samplefreq, int channels, int bits,
		 int samplenum)
{
  return sound_wave_open(filename, samplefreq, channels, bits, samplenum);
}

int sound_write(FILE *fp, int channels, int bits, int samplenum,
		int *offset, void *frames, int num)
{
  return sound_wave_write(fp, channels, bits, samplenum, offset, frames, num);
}

int sound_close(FILE *fp, int channels, int bits, int samplenum, int *offset)
{
  return sound_wave_close(fp, channels, bits, samplenum, offset);
}
